from dios.lib import *
from dios.options import *
from dios.itypes import *
from dios.errors import *
import pandas as pd
import numpy as np
import operator as op
from functools import partialmethod
from itertools import zip_longest, compress, takewhile

from functools import wraps
from collections import OrderedDict
from pandas._libs.lib import to_object_array, is_bool_array
from pandas.core.common import is_bool_indexer
from pandas.core.dtypes.common import (
    is_list_like,
    is_nested_list_like,
    is_scalar,
    is_integer,
    is_dict_like,
    is_number,
    is_hashable,
)
from pandas.core.dtypes.common import is_iterator as _is_iterator


def is_list_like_not_nested(obj):
    return is_list_like(obj) and not is_nested_list_like(obj)


def is_dios_like(obj):
    # must have columns
    # columns is some kind of pd.Index
    # iter will iter through columns
    # a `in` obj check if obj is in columns
    # obj[key] will give a pd.Series
    # obj.squeeze() give pd.Series if len(obj) == 1
    return isinstance(obj, DictOfSeries) or isinstance(obj, pd.DataFrame)


def is_bool_series(obj):
    return isinstance(obj, pd.Series) and obj.dtype == bool


def is_iterator(obj):
    """ This is only a dummy wrapper, to warn that the docu of this isnt't right.
    Unlike the example says,
    >>is_iterator([1, 2, 3])
    returns False, not True for lists
    """
    return _is_iterator(obj)


class DictOfSeries:
    """
    DictionaryOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned pd.Series.

    DictionaryOfSeries:
      - ``dios['a']``       indexing with single label, work on columns, return series
      - ``dios[iterable]``  indexing with list, work on columns, return dios
      - ``dios[slice]``     indexing with slice, work on (all) index'es, return dios
      - ``dios[bool_pdser]``indexing with a pd.Series with boolean values, work on (all) index'es, align, return dios
      - ``dios.loc[..]``    loc[cation] based indexing, work on row and/or columns `labels`, align, return dios
      - ``dios.iloc[..]``   i[nteger]loc[cation] based indexing, work on row and/or columns, align, return dios

    Todos:
    -----
    """

    def __init__(self, data=None, columns=None, itype=MixedItype, downcast_policy='save'):

        self._data = pd.Series(dtype='O')
        # We need to keep track of the index-type (itype) of every new Series.
        # If the itypes differ between different series, slicing will almost always fail
        # (eg. a datetime-like slice cannot work on a numeric index and vice versa).
        self._itype = get_itype(itype)
        if not self._itype.unique:
            throw_MixedItype_err_or_warn(self.itype)

        if downcast_policy not in CAST_POLICIES:
            raise ValueError(f"downcast_policy must be one of {CAST_POLICIES}")
        self._policy = downcast_policy

        if columns is not None and not is_list_like_not_nested(columns):
            raise TypeError("'columns' must be some kind of list-like collection.")

        if data is not None:
            self._init_insert_data(data, columns)

        # insert empty if requested
        if columns is not None:
            for c in columns:
                if c not in self.columns:
                    self._insert(c, pd.Series())

        for s in self._data:
            cast_to_itype(s, self.itype, policy=self._policy, inplace=True)

    def _init_insert_data(self, data, columns):
        def incols(c):
            return c in columns if columns is not None else True

        if isinstance(data, dict) or is_dios_like(data):
            for k in data:
                if incols(k):
                    self._insert(k, data[k])

        elif is_list_like(data):  # also Series !
            data = data if is_nested_list_like(data) else [data]
            if columns is None:
                for i, d in enumerate(data):
                    self._insert(i, d)
            else:
                if len(data) != len(columns):
                    raise ValueError(f"length of passed values is {len(data)}, columns imply {len(columns)}")
                for i, c in enumerate(columns):
                    self._insert(c, data[i])
        else:
            raise ValueError(f"Cannot insert data of type {type(data)}")

    def _insert(self, col, val):
        """Insert a fresh new value into self"""
        if is_dios_like(val):
            if len(val) > 1:
                raise ValueError(f"Cannot insert DictOfSeries "
                                 f"with more than one series")
            val = val.squeeze()
        else:
            val = pd.Series(val)

        if not isinstance(val, pd.Series):
            raise ValueError(f"Cannot insert data of type {type(val)}")

        val = cast_to_itype(val, self.itype, policy=self._policy)
        self._data.at[col] = val.copy(deep=True)

    @property
    def columns(self):
        return self._data.index

    @columns.setter
    def columns(self, newindex):
        self._data.index = newindex

    @property
    def itype(self):
        return self._itype

    @itype.setter
    def itype(self, newitype):
        itype = get_itype(newitype)

        if not itype_le(self._itype, itype):
            self.__cast_all(itype)

        self._itype = itype

        if not itype.unique:
            throw_MixedItype_err_or_warn(f"Using a {itype} as dios.itype is experimental. As soon as series with "
                                         f"different index types are inserted, slicing will almost always fail. "
                                         f"You are hereby warned!")

    def __cast_all(self, itype):
        c = '?'
        try:
            for c in self.columns:
                self[c] = cast_to_itype(self[c], itype, policy=self._policy)
        except Exception as e:
            raise type(e)(f"Column {c}: " + str(e)) from e

    def __getitem__(self, key):
        """ dios[key] -> dios/series """

        if isinstance(key, tuple):
            raise KeyError("tuples are not allowed")
        elif is_hashable(key):
            new = self._data.at[key]
        elif is_dios_like(key):
            # work on rows and columns
            new = self._getitem_bool_dios(key)
        elif isinstance(key, slice):
            # work on rows
            new = self._slice(key)
        elif is_bool_indexer(key):
            # work on rows
            new = self._getitem_bool_listlike(key)
        else:
            # work on columns
            new = self.copy_empty(columns=False)
            new._data = self._data.loc[key]
        return new

    def _slice(self, key):
        """slices self, return copy"""
        if key == slice(None):
            return self.copy()

        new = self.copy_empty(columns=False)
        for k in self.columns:
            # bug-fix: must be .loc, simple
            # ser[key] may work positional!
            new._data.at[k] = self._data.at[k].loc[key]
        return new

    def _getitem_bool_dios(self, key):
        # align columns
        keys = self.columns.intersection(key.columns)
        new = self.copy_empty(columns=True)

        for k in keys:
            ser = self._data.at[k]
            boolser = key[k]
            if not is_bool_indexer(boolser):
                raise ValueError("Must pass DictOfSeries with boolean values only")
            # align rows
            idx = boolser[boolser].index.intersection(ser.index)
            new._data.at[k] = ser.loc[idx]
        return new

    def _getitem_bool_listlike(self, key):
        new = self.copy_empty(columns=False)
        for k in self.columns:
            new._data.at[k] = self._data.at[k].loc[key]
        return new

    def __setitem__(self, key, value):
        """ dios[key] = value """

        if isinstance(key, tuple):
            raise KeyError(f"{key}. tuples are not allowed")

        elif is_hashable(key):
            if isinstance(value, pd.Series):
                self._insert(key, value)
            else:
                self._data.at[key][:] = value

        else:
            data = self.__getitem__(key)
            assert isinstance(data, self.__class__), f"getitem returned data of type {type(data)}"

            if is_dios_like(value):
                self._setitem_dios(data, value)
            else:
                for k in data.columns:
                    s = data._data.at[k]
                    s[:] = value
                    self._data.at[k].loc[s.index] = s

    def _setitem_dios(self, data, value):
        keys = self.columns.intersection(data.columns)
        for k in keys:
            l = self._data.at[k]
            r = value[k]
            idx = l.index.intersection(r.index)
            l[idx] = r[idx]

    @property
    def loc(self):
        from dios.locator import _LocIndexer
        return _LocIndexer(self)

    @property
    def iloc(self):
        from dios.locator import _iLocIndexer
        return _iLocIndexer(self)

    @property
    def aloc(self):
        from dios.locator import _aLocIndexer
        return _aLocIndexer(self)

    @property
    def at(self):
        from dios.locator import _AtIndexer
        return _AtIndexer(self)

    @property
    def iat(self):
        from dios.locator import _iAtIndexer
        return _iAtIndexer(self)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return pprint(self,
                      max_rows=dios_options[OptsFields.disp_max_rows],
                      max_cols=dios_options[OptsFields.disp_max_vars])

    @property
    def empty(self):
        return len(self) == 0 or all(s.empty for s in self._data)

    def all(self, axis=0):
        if axis in [0, 'index']:
            return self._data.apply(all)
        elif axis in [1, 'columns']:
            raise NotImplementedError
        elif axis is None:
            return self._data.apply(all).all()
        raise ValueError(axis)

    def any(self, axis=0):
        if axis in [0, 'index']:
            return self._data.apply(any)
        elif axis in [1, 'columns']:
            raise NotImplementedError
        elif axis is None:
            return self._data.apply(any).any()
        raise ValueError(axis)

    def squeeze(self, axis=None):
        if axis in [0, 'index']:
            raise NotImplementedError

        if len(self) > 1:
            return self

        if axis in [1, 'columns']:
            return self._data.squeeze()
        elif axis is None:
            return self._data.squeeze().squeeze()

        raise ValueError(axis)

    def __len__(self):
        return len(self.columns)

    def __iter__(self):
        yield from self.columns.copy()

    def __reversed__(self):
        yield from reversed(self.columns.copy())

    def __contains__(self, item):
        return item in self.columns.copy()

    def __delitem__(self, key):
        del self._data[key]

    def __copy__(self):
        return self.copy(deep=True)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def copy(self, deep=True):
        new = DictOfSeries(itype=self.itype)
        # We use `_data` here, because all checks are already done.
        # So this should be much faster, especially, because we use the underlying dict for
        # getting and setting the values, instead of ``__setitem__`` and ``__getitem__``.
        for i in self._data.index:
            new._data.at[i] = self._data.at[i].copy(deep=deep)
        return new

    def copy_empty(self, columns=True):
        return DictOfSeries(columns=self.columns if columns is True else None, itype=self.itype)

    def to_df(self):
        return self._data.apply(lambda s: s).transpose()

    @property
    def debugDf(self):
        return self.to_df()

    def memory_usage(self, index=True, deep=False):
        mem = 0
        for k in self.columns:
            mem += self[k].memory_usage(index=index, deep=deep)
        return mem

    def apply(self, func, axis=0, raw=False, args=(), **kwds):
        if axis in [1, 'columns']:
            raise NotImplementedError

        elif axis in [0, 'index']:
            pass
        else:
            raise ValueError(axis)

        # we cannot use self._data.apply(func=func, args=args, **kwds)
        # because this may return a df

        need_dios = False
        new = []
        for c in self.columns:
            s = func(self[c].values if raw else self[c], *args, **kwds)
            s = s.squeeze() if isinstance(s, pd.Series) else s
            new.append(s)
            try:
                if len(s) > 1:
                    need_dios = True
            except TypeError:
                pass

        if need_dios:
            dios = self.copy_empty(columns=False)
            for i, c in enumerate(self.columns):
                dios[c] = pd.Series(new[i])
            new = dios
        else:
            new = pd.Series(data=new, index=self.columns)
        return new

    @property
    def dtypes(self):
        s = pd.Series(index=self.columns, dtype='O')
        for k in self.columns:
            s.at[k] = self._data.at[k].dtype
        return s

    def astype(self, dtype, copy=True, errors='raise'):
        new = self.copy_empty(columns=False) if copy else self
        for k in self.columns.copy():
            new._data.at[k] = self._data.at[k].astype(dtype=dtype, copy=True, errors=errors)
        return new

    def _op1(self, op):
        new = self.copy_empty(columns=False)
        try:
            for k in self.columns:
                new[k] = op(self[k])
        except Exception as e:
            raise type(e)(f"'{OP_MAP[op]} dios' failed: " + str(e)) from e
        return new

    def _op2(self, op, other, align=True, inplace=False):
        def raiseif(kself, kother, s):
            if kself != kother:
                raise ValueError(f"{s} does not match, {s} left: {kself}, {s} right: {kother}")

        def doalign(left, right):
            return left.align(right, join='inner') if align else (left, right)

        def gen():
            if is_dios_like(other):
                raiseif(list(self), list(other), 'keys')
                for k in self.columns:
                    left, right = self[k], other[k]
                    yield k, op(*doalign(left, right))
            elif isinstance(other, pd.Series):
                for k in self.columns:
                    left, right = self[k], other
                    yield k, op(*doalign(left, right))
            elif is_dict_like(other):
                raiseif(sorted(self), sorted(other), 'keys')
                for k in self.columns:
                    yield k, op(self[k], other[k])
            elif is_nested_list_like(other):
                raiseif(len(self), len(other), 'length')
                for i, k in enumerate(self.columns):
                    yield k, op(self[k], other[i])
            elif is_scalar(other) or is_list_like(other):
                for k in self.columns:
                    yield k, op(self[k], other)
            else:
                raise NotImplementedError

        new = self if inplace else self.copy_empty(columns=True)
        try:
            for k, val in gen():
                new[k] = val
        except Exception as e:
            raise type(e)(f"'dios {OP_MAP[op]} other' failed: " + str(e)) from e
        return new

    __neg__ = partialmethod(_op1, op.neg)
    __abs__ = partialmethod(_op1, op.abs)
    __invert__ = partialmethod(_op1, op.inv)
    __eq__ = partialmethod(_op2, op.eq, align=False)
    __ne__ = partialmethod(_op2, op.ne, align=False)
    __le__ = partialmethod(_op2, op.le, align=False)
    __ge__ = partialmethod(_op2, op.ge, align=False)
    __lt__ = partialmethod(_op2, op.lt, align=False)
    __gt__ = partialmethod(_op2, op.gt, align=False)
    __add__ = partialmethod(_op2, op.add)
    __sub__ = partialmethod(_op2, op.sub)
    __mul__ = partialmethod(_op2, op.mul)
    __mod__ = partialmethod(_op2, op.mod)
    __truediv__ = partialmethod(_op2, op.truediv)
    __floordiv__ = partialmethod(_op2, op.floordiv)
    __pow__ = partialmethod(_op2, op.pow)
    __and__ = partialmethod(_op2, op.and_)
    __or__ = partialmethod(_op2, op.or_)
    __xor__ = partialmethod(_op2, op.xor)


def pprint(dios, max_rows=10, max_cols=2, delim='  '):
    if dios.empty:
        return f"Empty DictionaryOfSeries\n" \
               f"Columns: {dios.columns.to_list()}"

    sstr = []
    cols = list(dios.columns)

    for c in dios.columns:
        if dios[c].empty:
            sstr.append(['no data'])
        else:
            sstr.append(dios[c].to_string(max_rows=max_rows).split('\n'))

    maxlen = max([len(x) for x in sstr])

    # reduce number of visible vars
    if len(sstr) > max_cols:
        l, r = sstr[:max_cols // 2], sstr[-(max_cols // 2):]
        cols = cols[:max_cols // 2] + [" ..."] + cols[-(max_cols // 2):]
        maxlen = max([len(x) for x in l + r])
        l.append([" ..."] * maxlen)
        sstr = l + r

    # align inner and outer
    outer, header, subheader = [], [], []
    for j, s in enumerate(sstr):
        ml = max([len(x) for x in s])
        name = cols[j] + " " * ml
        header.append(name[:ml])
        subheader.append("-" * ml)
        inner = []
        for i in s:
            # make all inner lists same length
            inner.append(i + " " * (ml - len(i)))
        # make all outer lists same length
        outer.append(inner + [" " * ml] * (maxlen - len(s)))

    # switch outer and inner
    rstr = list(map(list, zip_longest(*outer, fillvalue="")))

    txt = ""
    rstr = [header] + [subheader] + rstr
    for r in rstr:
        for s in r:
            txt += s + delim
        txt += '\n'

    return txt
